/* File built by lookups/Makefile */
#include "../exim.h"
extern lookup_module_info dbmdb_lookup_module_info;
extern lookup_module_info dnsdb_lookup_module_info;
extern lookup_module_info dsearch_lookup_module_info;
extern lookup_module_info ldap_lookup_module_info;
extern lookup_module_info lsearch_lookup_module_info;
extern lookup_module_info sqlite_lookup_module_info;
extern lookup_module_info spf_lookup_module_info;
extern lookup_module_info readsock_lookup_module_info;
lookup_module_info * avail_static_lookups[] = {
& dbmdb_lookup_module_info,
& dnsdb_lookup_module_info,
& dsearch_lookup_module_info,
& ldap_lookup_module_info,
& lsearch_lookup_module_info,
& sqlite_lookup_module_info,
& spf_lookup_module_info,
& readsock_lookup_module_info,
NULL };
